import os
import sys

sys.path.append('/scratch/stimela')
import utils

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
args = {}
parset = []

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue
    elif value is False:
        value = 0
    elif value is True:
        value = 1
    elif name == 'parset':
       parset = [value]
       continue
    elif name == 'model-list':
        if isinstance(value, str):
            value = [value]
        for i,val in enumerate(value):
            if not os.path.exists(val.split("@")[0]):
                value[i] = os.path.basename(val)
        value = ':'.join(value)
    elif isinstance(value, list):
        value = ",".join( map(str, value) )

    args[name] = value

# available jones terms
joneses = "g b dd".split()
soljones = args.pop("sol-jones")

for jones in joneses:
    if jones.lower() not in soljones.lower():
        jopts = filter(lambda a: a.startswith("{0:s}-".format(jones)), args.keys())
        for item in jopts:
            del args[item]

opts =  ["{0:s}sol-jones {1:s}".format(cab["prefix"], soljones)] + \
        ['{0}{1} {2}'.format(cab['prefix'], name, value) for name,value in args.iteritems()]

utils.xrun(cab['binary'], parset+opts)



import sys
import numpy
import pyfits
import scipy.stats as stats



def get_imslice(ndim):
    imslice = []
    for i in xrange(ndim):
        if i<ndim-2:
            imslice.append(0)
        else:
            imslice.append(slice(None))

    return imslice


def fitsFreqInd(hdr):
    ndim = hdr["naxis"]
    keys = [hdr["CTYPE%d"%d].strip().lower() for d in xrange(1,ndim+1)]

    if "freq" not in keys:
        return False

    ind = keys.index("freq") + 1

    return ind


def freqInfo(fits):
    with pyfits.open(fits) as hdu:
        hdr = hdu[0].header
        ndim = hdr["NAXIS"]

    ind = fitsFreqInd(hdr)
    ref_freq = hdr["CRVAL%d"%ind]
    dfreq = hdr["CDELT%d"%ind]
    nchan = hdr["NAXIS%d"%ind]
    ref_chan = hdr["CRPIX%d"%ind]

    cnt_freq = ref_freq + (nchan/2 - ref_chan) * dfreq
    bw = dfreq*nchan

    freqs = numpy.linspace(cnt_freq-bw/2, cnt_freq+bw/2, nchan)

    return freqs, cnt_freq, bw, nchan, ind


def spifit(cube, prefix=None, mask=None, thresh=None,
           sigma=20, spi_image=None,
           spi_err_image=None):

    if isinstance(cube, (list, tuple)):
        ims = []
        nchan = len(cube)
        freqs = []
        prefix = prefix or cube[0][-3:]

        for im in cube:
            with pyfits.open(im) as hdu:
                hdr = hdu[0].header
                hdu_data = hdu[0].data
                ndim = hdr["NAXIS"]
                ind = fitsFreqInd(hdr)
                freqs.append( hdr["CRVAL%d"%ind])
                ims.append( hdu_data[get_imslice(ndim)]  )

        data = numpy.dstack(ims).T
        ndim = data.ndim
        cnt_freq = freqs[nchan/2]

    else:
        prefix = prefix or cube[-3:]
        with pyfits.open(cube) as hdu:
            data = hdu[0].data
            hdr = hdu[0].header
            ndim = data.ndim

        freqs, cnt_freq, bw, nchan, ind = freqInfo(cube)

    freq_ind = ndim - ind
    imslice = get_imslice(ndim)
    imslice[freq_ind] = slice(None)
    data = data[imslice]

    mfs = numpy.mean(data, axis=0)
    alpha = numpy.zeros(mfs.shape, dtype=numpy.float32)
    err = numpy.zeros(mfs.shape, dtype=numpy.float32)
    aa = []
    bb = []
    I0 = []

    if mask:
        with pyfits.open(mask) as hdu:
            mdata = hdu[0].data
            mndim = mdata.ndim
            mask = mdata[get_imslice(mndim)]

        ind = numpy.where(numpy.logical_and(mask>1e-33, mfs>1e-8))
    else:
        if thresh in [None, 0]:
            noise = numpy.median( abs(mfs - numpy.median(mfs))  )/0.667
            thresh = sigma*noise
        ind = numpy.where(mfs>thresh)

    if len(ind) < 1:
        raise RunTimeError("No pixels above set threshold, or outside masked region")

    for i,j in zip(ind[0], ind[1]):
        x = numpy.log(numpy.array(freqs)/cnt_freq)
        val = data[:,i,j]
        if val.any() <= 0:
            continue

        y = numpy.log( val )

        slope, intercept, r_value, p_value, std_err = stats.linregress(x,y)

        alpha[i,j] = slope
        err[i,j] = std_err

    #    aa.append(intercept)
    #    bb.append(std_err)
    #    I0.append(numpy.log(data[nchan/2,i,j]))

    nans = numpy.isnan(alpha)
    alpha[nans] = 0.0
    err[nans] = 1e99
    
    pyfits.writeto(spi_image or prefix+".alpha.fits", alpha, hdr, clobber=True)
    pyfits.writeto(spi_err_image or prefix+".alpha.err.fits", err, hdr, clobber=True)



import inner_taper as taper
import os
import sys

sys.path.append("/scratch/stimela")
import utils

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
params = {}

msname = None
res = None
freq = None
reset = False
savefig = None

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if name == 'min-max-resolution':
        res = value
    elif name == 'frequency':
        freq = value
    elif name == 'reset':
        reset = value
    elif name == 'msname':
        msname = value
    elif name == 'save-figure':
        savefig = value
        

if reset:
    taper.reset(msname)
    sys.exit(0)

taper.taper(msname, res=res, freq=freq, savefig=savefig)




#!/usr/bin/env python
import pyfits
import numpy as np
import Tigger
import sys,os
from astLib.astWCS import WCS
from scipy import ndimage

def fitsInfo(fitsname = None):
    """
    Get fits info
    """
    hdu = pyfits.open(fitsname)
    hdr = hdu[0].header
    ra = hdr['CRVAL1']
    dra = abs(hdr['CDELT1'])
    raPix = hdr['CRPIX1']
    dec = hdr['CRVAL2']
    ddec = abs(hdr['CDELT2'])
    decPix = hdr['CRPIX2']
    freq0 = 0
    for i in range(1,hdr['NAXIS']+1):
        if hdr['CTYPE%d'%i].strip() == 'FREQ':
            freq0 = hdr['CRVAL%d'%i]
            break

    ndim = hdr["NAXIS"]
    imslice = np.zeros(ndim, dtype=int).tolist()
    imslice[-2:] = slice(None), slice(None)
    image = hdu[0].data[imslice]
    wcs = WCS(hdr,mode='pyfits')

    return {'image':image,'wcs':wcs,'ra':ra,'dec':dec,'dra':dra,'ddec':ddec,'raPix':raPix,'decPix':decPix,'freq0':freq0}

def sky2px(wcs,ra,dec,dra,ddec,cell, beam):
    """convert a sky region to pixel positions"""
    dra = beam if dra<beam else dra # assume every source is at least as large as the psf
    ddec = beam if ddec<beam else ddec
    offsetDec = int((ddec/2.)/cell)
    offsetRA = int((dra/2.)/cell)
    if offsetDec%2==1:
        offsetDec += 1
    if offsetRA%2==1:
        offsetRA += 1

    raPix,decPix = map(int, wcs.wcs2pix(ra,dec))
    return np.array([raPix-offsetRA,raPix+offsetRA,decPix-offsetDec,decPix+offsetDec])

def RemoveSourcesWithoutSPI(lsmname_in, lsmname_out):
    model = Tigger.load(lsmname_in)
    sources = [src for src in model.sources]
    for src in sources:
        if not src.spectrum:
            model.sources.remove(src)
    model.save(lsmname_out)

def CombineSourcesInCluster(lsmname_in, lsmname_out):
    model = Tigger.load(lsmname_in)
    for src in model.sources:
        if src.cluster_size>1 and rad2arcsec(src.r)>30:
            cluster_sources = [src1 for src1 in model.sources if src1.cluster is src.cluster]
            flux_sources = [src1.flux.I for src1 in cluster_sources]
            max_flux_index = flux_sources.index(max( flux_sources))
            cluster_sources[max_flux_index].flux.I = sum([src1.flux.I for src1 in cluster_sources])
        for src2 in cluster_sources:
            if src2 is not cluster_sources[max_flux_index]:
                model.sources.remove(src2)
        cluster_sources[max_flux_index].cluster_size=1
    model.save(lsmname_out)

def rad2arcsec(x):
    return x*3600.0*180.0/np.pi


def addSPI(fitsname_alpha=None, fitsname_alpha_error=None, lsmname=None,
           outfile=None,freq0=None, beam=None, spitol=(-10,10)):
    """
        Add spectral index to a tigger lsm from a spectral index map (fits format)
        takes in a spectral index map, input lsm and output lsm name.
    """
#    import pylab as plt
    if beam is None:
        raise RuntimeError("the beam option must be specified")

    print "INFO: Getting fits info from: %s, %s" %(fitsname_alpha, fitsname_alpha_error)

    fits_alpha = fitsInfo(fitsname_alpha)    # Get fits info
    image_alpha = fits_alpha['image']     # get image data

    if fitsname_alpha_error:
        fits_alpha_error = fitsInfo(fitsname_alpha_error)
        image_alpha_error = fits_alpha_error['image']
    else:
        fits_alpha_error = fitsInfo(fitsname_alpha)
        image_alpha_error = fits_alpha_error['image']
        image_alpha_error[...] = 1.0

    # may supply FITS file for freq0, in which case just pull ref frequency from FITS file,
    # else explicit frequency, else get frequency from alpha image
    if type(freq0) is str:
        freq0 = fitsInfo(freq0)['freq0']
    else:
        freq0 = freq0 or fits_alpha['freq0']

    model = Tigger.load(lsmname)    # load output sky model
    rad = lambda a: a*(180/np.pi) # convert radians to degrees

    for src in model.sources:
        ra = rad(src.pos.ra)
        dec = rad(src.pos.dec)

        # Cater for point sources and assume source extent equal to the
        # Gaussian major axis along both ra and dec axis
        dra = rad(src.shape.ex) if src.shape  else beam[0]
        ddec = rad(src.shape.ey) if src.shape  else beam[1]
        pa = rad(src.shape.pa) if src.shape  else beam[2]

        emin, emaj = sorted([dra, ddec])
        # Determine region of interest
        rgn = sky2px(fits_alpha["wcs"],ra,dec,dra,ddec,fits_alpha["dra"], beam[1])
        imslice = slice(rgn[2], rgn[3]), slice(rgn[0], rgn[3])
        alpha = image_alpha[imslice]
        xpix, ypix = alpha.shape
        xx, yy = np.ogrid[-xpix:xpix, -ypix:ypix]

        emajPix = emaj/fits_alpha["dra"]
        eminPix = emin/fits_alpha["dra"]
        # Create elliptcal mask which has same shape as source
        mask = ((xx/emajPix)**2 + (yy/eminPix)**2 <= 1)[xpix*2-xpix:xpix*2+xpix, ypix*2-ypix:ypix*2+ypix]
        mask = ndimage.rotate(mask, angle=pa, order=0, reshape=False)
        
        draPix = dra/fits_alpha["dra"]
        ddPix = ddec/fits_alpha["ddec"]

        alpha *= mask
        alpha_error = image_alpha_error[imslice]*mask
        good = np.where( np.logical_and(alpha!=0, alpha!=np.nan))
        alpha = alpha[good]
        alpha_error = alpha_error[good]
        good = np.where( np.logical_and(alpha_error!=np.nan, alpha_error!=np.inf))

        alpha = alpha[good]
        alpha_error = alpha_error[good]

        subIm_weight = 1/alpha_error
        subIm_weighted = alpha*subIm_weight

        if len(subIm_weighted)>0:
            subIm_normalization = np.sum(subIm_weight)
            spi = float(np.sum(subIm_weighted)/subIm_normalization)
            spi_error = 1/float(subIm_normalization)
            if spi > spitol[0] or spi < spitol[-1]:
                sys.stdout.write("INFO: Adding spi: %.3f (at %.3g MHz) to source %s" % (
                                 spi, freq0/1e6, src.name))
                src.spectrum = Tigger.Models.ModelClasses.SpectralIndex(spi, freq0)
                src.setAttribute('spi_error', spi_error)
        else:
            sys.stdout.write("ALERT: no spi info found in %s for source %s" % (
                             fitsname_alpha, src.name))

    model.save(outfile)

if __name__=="__main__":
    fitsname_alpha = sys.argv[1]
    #fitsname_alpha_error = sys.argv[2]
    lsmname = sys.argv[2]
    outfile = sys.argv[3]
    addSPI(fitsname_alpha, None, lsmname, outfile)



from stimela import utils
import logging
import sys
import os
import textwrap

USER = os.environ['USER']

TYPES = {
    "str"   :   str,
    "float" :   float,
    "bool"  :   bool,
    "int"   :   int,
    "list"  :   list,
   }


IODEST = {
    "input"     :   "/input",
    "output"    :   "/home/{}/output".format(USER),
    "msfile"    :   "/home/{}/msdir".format(USER),
}


class Parameter(object):
    def __init__(self, name, dtype, info, 
        default=False, 
        required=False,
        choices=None, 
        io=None,
        mapping=None,
        #delimiter=None, 
        check_io=True):

        self.name = name
        self.io = io
        #self.delimiter = delimiter
        
        if isinstance(dtype, str):
            dtype = [dtype]
        self.dtype = []
        for item in dtype:
            tmp = self.get_type(item)
            self.dtype.append(tmp)

        self.info = info
        self.default = default
        self.required = required
        self.choices = choices or []
        self.mapping = mapping
        self.check_io = check_io

        self.value = None

    def validate(self, value):
        if self.choices and value not in self.choices:
            raise ValueError("Parameter '{0}', can only be either of {1}".format(self.name, self.choices))

        for item in self.dtype:
            if isinstance(item, tuple):
                l,t = item
                if t=="file":
                    return True
                if isinstance(value, t):
                    return True
                elif isinstance(value, list):
                    if value == []:
                        return True
                    elif isinstance(value[0], tuple([t]+[int] if t is float else [t])):
                        return True
            elif item is "file":
                return True
            elif isinstance(value, tuple([item]+[int] if item is float else [item])):
                return True
        raise TypeError("Expecting any of types {0} for parameter '{1}', but got '{2}'".format( self.dtype, self.name, type(value).__name__))


    def get_type(self, dtype):

        def _type(a):
            if a == "file":
                if self.io not in ["input", "output", "msfile"]:
                    raise TypeError("io '{0}' for parameter '{1}' not understood. Please specify 'io' as either 'input', 'output' or 'msfile'".format(self.io, self.name))
                return "file"
            else:
                return TYPES[a]

        if dtype.startswith("list:"):
            val = dtype.split(":")
            if len(val) != 2:
                raise TypeError("The type of '{0}' could not validate. Specify list types as \"list:dtype\" where dtype is normal type")
            ttype = val[1]

            return (list, _type(ttype))
        else:
            return _type(dtype)
            


class CabDefinition(object):
    def __init__(self, 
        indir=None,  # input directory
        outdir=None, # output directory
        msdir=None,  # MS directory
        parameter_file=None,
        task=None,
        base=None,
        binary=None,
        description=None,
        tag=None,
        prefix=None, loglevel='INFO',
        parameters=[]):

        logging.basicConfig(level=getattr(logging, loglevel))
        self.log = logging
        self.indir = indir
        self.outdir = outdir

        if parameter_file:
            cab = utils.readJson(parameter_file)
            self.task = cab["task"]
            self.base = cab["base"]
            self.binary = cab["binary"]
            self.tag = cab["tag"]
            if cab["msdir"]:
                self.msdir = msdir
            self.description = cab["description"]
            self.prefix = cab["prefix"]
            parameters0 = cab["parameters"]
            self.parameters = []
            
            import sys
            for param in parameters0:
                default = param.get("default", param.get("value", None))
                addme = Parameter(name=param["name"],
                        dtype=param["dtype"],
                        io=param.get("io", None),
                        info=param.get("info", None) or "No documentation. Bad! Very bad...",
                        default=default,
                        mapping=param.get("mapping", None),
                        #delimiter=param.get("delimiter", None),
                        required=param.get("required", False),
                        choices=param.get("choices", False),
                        check_io=param.get("check_io", True))
                self.parameters.append(addme)

        else:
            self.task = task
            self.base = base
            self.binary = binary
            self.prefix = prefix
            self.parameters = parameters
            self.description = description
            self.msdir = msdir
            self.tag = tag

        
    def display(self, header=False):
        rows, cols = os.popen('stty size', 'r').read().split()
        lines = textwrap.wrap(self.description, int(cols)*3/4)
        print("Cab      {0}".format(self.task))
        print("Info     {}".format(lines[0]))
        for line in lines[1:]:
            print("         {}".format(line))
        if header:
            print(" ")
            return

        print("Base Image       {0}:{1}".format(self.base, self.tag))
        print("\n")

        print("Parameters:")
        rows, cols = os.popen('stty size', 'r').read().split()
        for param in self.parameters:
            
            _types = ""
            for i,_type in enumerate(param.dtype):
                if isinstance(_type, tuple):
                    _name = "list:{}".format("file" if _type[1]=="file" else _type[1].__name__)
                else:
                    _name = "file" if _type=="file" else _type.__name__
                _types += "{}".format(_name) if i==0 else "/{}".format(_name)

            lines = textwrap.wrap(param.info, int(cols)*3/4)

            print("  Name         {}{}".format(param.name, "/{}".format(param.mapping) if param.mapping else ""))
            print("  Description  {}".format(lines[0]))
            for line in lines[1:]:
                print("               {}".format(line))
            print("  Type         {}".format(_types))
            print("  Default      {}".format(param.default))
            if param.choices:
                print("  Choices      {}".format(param.choices))
            print(" ")


    def toDict(self):
        conf = {}
        for item in "task base binary msdir description prefix tag".split():
            if item == 'msdir':
                conf[item] = getattr(self, item, False)
            else:
                conf[item] = getattr(self, item)
        
        conf["parameters"] = []
        for param in self.parameters:
        
            if isinstance(param.dtype[0], tuple):
                if not hasattr(param.value, '__iter__') and param.value is not None:
                    param.value = [param.value]

            _types = ""
            for i,_type in enumerate(param.dtype):
                if isinstance(_type, tuple):
                    _name = "list:{}".format("file" if _type[1]=="file" else _type[1].__name__)
                else:
                    _name = "file" if _type=="file" else _type.__name__
                _types += "{}".format(_name) if i==0 else "/{}".format(_name)

            conf["parameters"].append(
                {
                    "name"      :   param.mapping or param.name,
                    "dtype"     :   _types,
                    "info"      :   param.info,
                    "required"  :   param.required,
                    "check_io"  :   param.check_io,
                    "value"     :   param.default if param.value is None else param.value
                })
        return conf


    def update(self, options, saveconf):
        required = filter(lambda a: a.required, self.parameters)
        for param0 in required:
            if param0.name in options.keys() == False and param0.mapping in options.keys() == False:
                raise RuntimeError("Parameter {} is required but has not been specified".format(param0.name))

        self.log.info("Validating parameters...       CAB = {0}".format(self.task))
        for name,value in options.items():
            found = False
            for param in self.parameters:
                if name in [param.name, param.mapping]:
                    found = True
                    if param.io:
                        if value is None:
                            continue
                        param.validate(value)
                        param.value = []
                        if hasattr(value, "__iter__") and not isinstance(value, str):
                            print(value)
#                            pass
                        else:
                            value = [value]
                            print(value)
                        for _value in value:
                            val = _value.split(":")
                            if len(val)==2:
                                if val[1] not in IODEST.keys():
                                    raise IOError('The location \'{0}\' specified for parameter \'{1}\', is unknown. Choices are {2}'.format(val[1], param.name, IODEST.keys()))
                                self.log.info("Location of '{0}' was specified as '{1}'. Will overide default.".format(param.name, val[1]))
                                _value = val[0]
                                location = val[1]
                            else:
                                location = param.io

                            if location in ["input", "msfile"]:
                                if location == "input" and self.indir is None:
                                    raise IOError("You have specified input files, but have not specified an input folder")
                                if location == "msfile" and self.msdir is None:
                                    raise IOError("You have specified MS files, but have not specified an MS folder")

                                path = "{0}/{1}".format(self.indir if location=="input" else self.msdir, _value)
                                if param.check_io and not os.path.exists(path):
                                    raise IOError("File '{0}' for parameter '{1}' could not be located at '{2}'.".format(_value, param.name, path))
                                param.value.append ("{0}/{1}".format(IODEST[location], _value))
                            else:
                                if self.outdir is None:
                                    raise IOError("You have specified output files, but have not specified an output folder")
                                param.value.append("{0}/{1}".format(IODEST[location], _value))
                        if len(param.value)==1:
                            param.value = param.value[0]
                        
                    else:
                        self.log.debug("Validating paramter {}".format(param.name))
                        param.validate(value)
                        param.value = value
            if not found:
                raise RuntimeError("Parameter {0} is unknown. Run 'stimela cabs -i {1}' to get help on this cab".format(name, self.task))
        conf = {}
        conf.update(self.toDict())
        utils.writeJson(saveconf, conf)
        self.log.info("Parameters validated and saved. Parameter file is: {}".format(saveconf))



import pyrap.images
import os 
import sys
from pyrap.tables import table
from MSUtils import msutils
import tempfile
import pyfits

sys.path.append('/scratch/stimela')
import utils

CONFIG = os.environ['CONFIG']
OUTPUT = os.environ['OUTPUT']

cab = utils.readJson(CONFIG)
params = cab['parameters']

tdir = tempfile.mkdtemp(dir='.')
os.chdir(tdir)

def rm_fr(item):
    os.system('rm -fr {}'.format(item))

def _run (prefix=None, predict=False, **kw):

    if predict:
        args = [ '{0}={1}'.format(a, b) for a,b in kw.iteritems() ]
        utils.xrun(cab['binary'], args)
        return

    if kw.get('niter', 0) >0:
        if kw.get('operation', None) not in ['clark', 'hogbom', 'csclean', 'multiscale', 'entropy']:
            kw['operation'] = 'csclean'
        images = {
            "restored"  :   [ '{0}.restored.{1}'.format(prefix, a) for a in ['fits', 'img']],
            "model"  :   [ '{0}.model.{1}'.format(prefix, a) for a in ['fits', 'img']],
            "residual"  :   [ '{0}.residual.{1}'.format(prefix, a) for a in ['fits', 'img']],
        }

    elif kw.get('niter', 0) == 0:
        kw["operation"] = 'image'

        images = {
        "image"  :   [ '{0}.dirty.{1}'.format(prefix, a) for a in ['fits', 'img']],
        }
    
    for key,value in images.iteritems():
        kw[key] = value[1]
    
    args = [ '{0}={1}'.format(a, b) for a,b in kw.iteritems() ]
    utils.xrun(cab['binary'], args)

    print('Converting CASA iamges to FITS images')
    for fits,img in images.itervalues():
        im = pyrap.images.image(img)
        im.tofits(fits, overwrite=True, velocity=kw.get('prefervelocity', False))
        rm_fr(img)



def predict_vis (msname, image, column="MODEL_DATA", 
        chanchunk=None, chanstart=0, chanstep=1):
  """Converts image into predicted visibilities"""

  # CASA to convert them
  casaimage = '{0}/{1}.img'.format(OUTPUT, os.path.basename(image))
  
  # convert to CASA image
  img = pyrap.images.image(image)
  img.saveas(casaimage)

  imgshp = img.shape()
  ftab = table(msname+'/SPECTRAL_WINDOW')
  numchans = ftab.getcol('NUM_CHAN')[0]
  # default chunk list is entire chanel range. Update this if needed
  chunklist = [ (0, numchans, None, None) ]
  if len(imgshp) == 4 and imgshp[0] > 1:
      nimgchan = imgshp[0]

      print("image cube has {0} channels, MS has {1} channels".format(nimgchan, numchans))

      imgchansize = imgshp[1]*imgshp[2]*imgshp[3]*4  # size of an image channel in bytes
      if chanchunk is None:
          mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')  # e.g. 4015976448
          chanchunk = max((mem_bytes/20)/imgchansize, 1)
          print("based on available memory ({0}), max image chunk is {1} channels".format(mem_bytes, chanchunk))
      if chanchunk < nimgchan:
          mschanstep = numchans*chanstep/nimgchan
          if numchans%nimgchan:
              warn("MS channels not evenly divisible into $nimgchan image channels, chunking may be incorrect")
          chunklist = []
          for chan0 in range(0, nimgchan, chanchunk):
              imch0, imch1 = chan0, (min(chan0+chanchunk, nimgchan)-1)
              msch0 = chanstart + imch0*mschanstep
              msnch = (imch1-imch0+1)*mschanstep/chanstep
              # overlap each chunk from 1 onwards by a half-chunk back to take care of extrapolated visibilties
              # from previous channel
              if imch0:
                  imch0 -= 1
                  msch0 -= mschanstep/2
                  msnch += mschanstep/2
              print("image chunk {0}~{1} corresponds to MS chunk {2}~{3}".format(imch0, imch1, msch0, msch0+msnch-1))
              chunklist.append((msch0, msnch, imch0, imch1))
              
  # even in fill-model mode where it claims to ignore image parameters, the image channelization
  # arguments need to be "just so" as per below, otherwise it gives a GridFT: weights all zero message
  kw0 = {}
  kw0.update(ms=msname, model=casaimage, 
      niter=0, fixed=1, mode="channel", operation="csclean",
      img_nchan=1, img_chanstart=chanstart, img_chanstep=numchans*chanstep)
  kw0['fillmodel'] = 1
                   
  blc = [0]*len(imgshp)
  trc = [ x-1 for x in imgshp ]

  # now loop over image frequency chunks
  for ichunk, (mschanstart, msnumchans, imgch0, imgch1) in enumerate(chunklist):
      if len(chunklist) > 1:
          blc[0], trc[0] = imgch0, imgch1
          print("writing CASA image for slice {0} {1}".format(blc, trc))
          casaimage1 = II("{0}.{1}.img".format(image, ichunk1))
          rm_fr(casaimage1)
          print("writing CASA image for slice {0} {1} to {2}".format(blc, trc, casaimage1))
          img.subimage(blc, trc, dropdegenerate=False).saveas(casaimage1)
          kw0.update(model=casaimage1)
      else:
          img.unlock()
      # setup imager options
      kw0.update(chanstart=mschanstart, chanstep=chanstep, nchan=msnumchans)
      print("predicting visibilities into MODEL_DATA")

      _run(predict=True, **kw0)
      if len(chunklist) > 1:
        rm_fr(casaimage1)   
  rm_fr(casaimage)
  
  if column != "MODEL_DATA":
    print('Data was predicted to MODEL_DATA column. Will now copy it to the {} column as requested'.format(column))
    msutils.copycol(msname=msname, fromcol="MODEL_DATA", tocol=column)


options = {}
for param in params:
    value = param['value']
    name = param['name']

    if name == 'prefix':
        prefix = value
        continue

    if value is None:
        continue
    if name == 'cellsize':
        if isinstance(value, (float, int)):
            value = '{}arcsec'.format(value)
    elif name in ['threshold', 'targetflux']:
        if isinstance(value, float):
            value = '{}arcsec'.format(value)
    options[name] = value

noise_image = options.pop('noise_image', False)
if noise_image:
    noise_sigma = options.pop('noise_sigma')
    noise_hdu = pyfits.open(noise_image)
    noise_data = noise_hdu[0].data
    noise_std = noise_data.std()
    threshold = noise_sigma*noise_std
    options['threshold'] = '{}Jy'.format(threshold)
else:
    options.pop('noise_sigma')

predict = options.pop('simulate_fits', False)
if predict:
    tfile = tempfile.NamedTemporaryFile(suffix='.fits')
    tfile.flush()
    cell = options.get('cellsize', None)
    if cell is None:
        with pyfits.open(predict) as _hdu:
            if hasattr(_hdu, '__iter__'):
                hdu = _hdu[0]
            else:
                hdu = _hdu

            cdelt = hdu.header.get('CDELT1', None)
            if cdelt:
                cell = '{:f}arcsec'.format(abs(cdelt)*3600)

    if cell is None:
        raise RuntimeError('The size of a pixel in this FITS image was not specified \
in FITS header (CDELT1/2), or as parameter for this module ("cellsize"). Cannot proceed')
    
    utils.xrun('python /scratch/code/predict_from_fits.py', [predict, options['ms'], cell,
                tfile.name])
    predict_vis(msname=options['ms'], image=tfile.name, column=options.get('data','MODEL_DATA'), 
        chanchunk=options.get('chanchunk', None), chanstart=options.get('img_chanstart', 0), 
        chanstep=options.get('img_chanstep', 1))
    tfile.close()
else:
    _run(prefix, **options)

os.chdir(OUTPUT)
os.system('rm -r {}'.format(tdir))



import os
import sys
import re
import bdsf as bdsm #bdsm it is and bdsm it shall remain
import numpy
import Tigger
import tempfile
import pyfits

from astLib.astWCS import WCS
from Tigger.Models import SkyModel, ModelClasses


sys.path.append('/scratch/stimela')
import utils

CONFIG = os.environ['CONFIG']
INPUT = os.environ['INPUT']
OUTPUT = os.environ['OUTPUT']
MSDIR = os.environ['MSDIR']

cab = utils.readJson(CONFIG)

write_catalog = ['bbs_patches', 'bbs_patches_mask',
            'catalog_type', 'clobber', 'correct_proj', 'format',
            'incl_chan', 'incl_empty', 'srcroot', 'port2tigger', 'outfile']

img_opts = {}
write_opts = {}
# Spectral fitting parameters
freq0 = None
spi_do = False

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue
    if name in ['multi_chan_beam']:
        multi_chan_beam = value
        continue
    if name in write_catalog:
        write_opts[name] = value
    elif name in ['freq0', 'frequency']:
        freq0 = value
    else:
        img_opts[name] = value
        if name == 'spectralindex_do':
            spi_do = value

img_opts.pop('freq0', None)
if freq0 is None:
    with pyfits.open(img_opts['filename']) as hdu:
        hdr = hdu[0].header
        for i in xrange(1, hdr['NAXIS']+1):
            if hdr['CTYPE{0:d}'.format(i)].startswith('FREQ'):
                freq0 = hdr['CRVAL{0:d}'.format(i)]

if spi_do and multi_chan_beam:
    with pyfits.open(img_opts['filename']) as hdu:
        hdr = hdu[0].header
    beams = []
    # Get a sequence of BMAJ with digit suffix from the image header keys
    bmaj_ind = filter(lambda a: a.startswith('BMAJ') and a[-1].isdigit(), hdr.keys())
    for bmaj in bmaj_ind:
        ind = bmaj.split('BMAJ')[-1]
        beam = [hdr['{0:s}{1:s}'.format(b, ind)] for b in 'BMAJ BMIN BPA'.split()]
        beams.append(tuple(beam))
    # parse beam info to pybdsm
    img_opts['beam_spectrum'] = beams

image = img_opts.pop('filename')
#img_opts["indir"] = os.path.dirname(image) + "/"
filename = os.path.basename(image)
 
outfile = write_opts.pop('outfile')

img = bdsm.process_image(image, **img_opts)

port2tigger = write_opts.pop('port2tigger', True)

if port2tigger:
    write_opts['format'] = 'fits'

img.write_catalog(outfile=outfile, **write_opts)

if not port2tigger:
    sys.exit(0)

# convert to Gaul file to Tigger LSM
# First make dummy tigger model
tfile = tempfile.NamedTemporaryFile(suffix='.txt')
tfile.flush()

prefix = os.path.splitext(outfile)[0]
tname_lsm = prefix + ".lsm.html"
with open(tfile.name, "w") as stdw:
    stdw.write("#format:name ra_d dec_d i emaj_s emin_s pa_d\n")

model = Tigger.load(tfile.name)
tfile.close()

def tigger_src(src, idx):

    name = "SRC%d"%idx

    flux = ModelClasses.Polarization(src["Total_flux"], 0, 0, 0, I_err=src["E_Total_flux"])
    ra, ra_err = map(numpy.deg2rad, (src["RA"], src["E_RA"]) )
    dec, dec_err = map(numpy.deg2rad, (src["DEC"], src["E_DEC"]) )
    pos = ModelClasses.Position(ra, dec, ra_err=ra_err, dec_err=dec_err)
    ex, ex_err = map(numpy.deg2rad, (src["DC_Maj"], src["E_DC_Maj"]) )
    ey, ey_err = map(numpy.deg2rad, (src["DC_Min"], src["E_DC_Min"]) )
    pa, pa_err = map(numpy.deg2rad, (src["PA"], src["E_PA"]) )

    if ex and ey:
        shape = ModelClasses.Gaussian(ex, ey, pa, ex_err=ex_err, ey_err=ey_err, pa_err=pa_err)
    else:
        shape = None
    source = SkyModel.Source(name, pos, flux, shape=shape)
    # Adding source peak flux (error) as extra flux attributes for sources,
    # and to avoid null values for point sources I_peak = src["Total_flux"]
    if shape:
        source.setAttribute("I_peak", src["Peak_flux"])
        source.setAttribute("I_peak_err", src["E_peak_flux"])
    else:
        source.setAttribute("I_peak", src["Total_flux"])
        source.setAttribute("I_peak_err", src["E_Total_flux"])

    if spi_do:
        # Check if start frequency is provided if not provided raise error.
        # It is used to define tigger source spectrum index frequency
        if freq0:
            spi, spi_err = (src['Spec_Indx'], src['E_Spec_Indx'])
            source.spectrum = ModelClasses.SpectralIndex(spi, freq0)
            source.setAttribute('spi_error', spi_err)
        else:
            raise RuntimeError("No start frequency (freq0) provided.")
    return source


with pyfits.open(outfile) as hdu:
    data = hdu[1].data

    for i, src in enumerate(data):
        model.sources.append(tigger_src(src, i))

wcs = WCS(image)
centre = wcs.getCentreWCSCoords()
model.ra0, model.dec0 = map(numpy.deg2rad, centre)

model.save(tname_lsm)
# Rename using CORPAT
utils.xrun("tigger-convert", [tname_lsm, "--rename -f"])



import os
import sys
import argparse
import time
from  argparse import ArgumentParser
import textwrap as _textwrap
import tempfile
import signal

import inspect
import stimela
from stimela import utils, cargo
from stimela.cargo import cab
from stimela.recipe import Recipe as Pipeline
from stimela.recipe import Recipe, PipelineException
from stimela import docker, singularity, udocker
import pkg_resources
import warnings

try:
    __version__ = pkg_resources.require("stimela")[0].version
except pkg_resources.DistributionNotFound:
    __version__ = "dev"

from stimela.utils import logger

# Get to know user
USER = os.environ["USER"]
UID = os.getuid()
GID = os.getgid()

# Set up logging infrastructure
LOG_HOME = os.path.expanduser("~/.stimela")
# This is is the default log file. It logs stimela images, containers and processes
LOG_FILE = "{0:s}/stimela_logfile.json".format(LOG_HOME)
#LOG_FILE = "{0:s}/{1:s}_stimela_logfile.json".format(LOG_HOME, USER)


# Get base images
# All base images must be on dockerhub
BASE = os.listdir(cargo.BASE_PATH)

# Get package cab images (user can add their own cab images)
# All package cab images must be based on stimela base images.

CAB = list()
    
for item in os.listdir(cargo.CAB_PATH):
    try: 
        # These files must exist for a cab image to be valid
        ls_cabdir = os.listdir('{0}/{1}'.format(cargo.CAB_PATH, item))
        dockerfile = 'Dockerfile' in ls_cabdir
        paramfile = 'parameters.json' in ls_cabdir
        srcdir = 'src' in ls_cabdir
    except OSError:
        continue
    if dockerfile and paramfile and srcdir:
        CAB.append(item)

GLOBALS = {'foo' : 'bar'}
del GLOBALS['foo']


class MultilineFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(' ', text).strip()
        paragraphs = text.split('|n ')
        multiline_text = ''
        for paragraph in paragraphs:
            formatted_paragraph = _textwrap.fill(paragraph, width, initial_indent=indent, subsequent_indent=indent) + '\n\n'
            multiline_text = multiline_text + formatted_paragraph
        return multiline_text

def register_globals():
    frame = inspect.currentframe().f_back
    frame.f_globals.update(GLOBALS)


def build(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv = ' ' + arg

    parser = ArgumentParser(description='Build executor (a.k.a cab) images')
    parser.add_argument("-b", "--base", action="store_true",
            help="Build base images")

    parser.add_argument("-c", "--cab", metavar="CAB,CAB_DIR",
            help="Executor image (name) name, location of executor image files")

    parser.add_argument("-uo", "--us-only",
            help="Only build these cabs. Comma separated cab names")

    parser.add_argument("-i", "--ignore-cabs", default="",
            help="Comma separated cabs (executor images) to ignore.")

    parser.add_argument("-nc", "--no-cache", action="store_true",
            help="Do not use cache when building the image")

    parser.add_argument("-bl", "--build-label", default=USER,
            help="Label for cab images. All cab images will be named <CAB_LABEL>_<cab name>. The default is $USER")

    args = parser.parse_args(argv)
    log = logger.StimelaLogger('{0:s}/{1:s}_stimela_logfile.json'.format(LOG_HOME, args.build_label), jtype="docker")

    if args.base:
        # Build base and meqtrees images first
        BASE.remove("base")
        BASE.remove("meqtrees")
        BASE.remove("casa")
        
        for image in ["base", "meqtrees", "casa"] + BASE:
            dockerfile = "{:s}/{:s}".format(cargo.BASE_PATH, image)
            image = "stimela/{0}:{1}".format(image, __version__)
            docker.build(image,
                         dockerfile)

        log.log_image(image, dockerfile, replace=True)
        log.write()

        return 0

    workdir = "/home/{}/output/".format(USER)
    build_args = [
                  "WORKDIR {:s}".format(workdir),
                  "RUN useradd -r -u {0:d} -U {1:s}".format(UID, USER),
                  "USER {0:s}".format(USER),
                  ]

    no_cache = ["--no-cache"] if args.no_cache else []


    if args.cab:
        cab_args = args.cab.split(",")

        if len(cab_args)==2:
            cab, path = cab_args
        else:
            raise ValueError("Not enough arguments for build command.")

        image = "{:s}_cab/{:s}".format(args.build_label, cab)

        docker.build(image,
                     path,
                     build_args=build_args, args=no_cache)

        log.log_image(image, path, replace=True, cab=True)
        log.write()
        return

    # Cabs and their locations   
    cabs = []
    dockerfiles = []

    if args.us_only:
        CABS = args.us_only.split(',')
    else:
        # Images that have been logged
        # This is crucial for making custom cabs
        logged_images = log.read().get('images', {})
        for key,val in logged_images.items():
            if val['CAB']:
                cabs.append(key)
                dockerfiles.append(val['DIR'])
        # If user wants to ignore some cabs
        IGNORE = args.ignore_cabs.split(",")
        CABS = set(CAB).difference(set(IGNORE))

    # Prioritise package images over logged images
    cabs = ["{:s}_cab/{:s}".format(args.build_label, cab) for cab in CABS] + cabs
    dockerfiles = [ "{:s}/{:s}".format(cargo.CAB_PATH, cab) for cab in CABS] + dockerfiles
    built = []
    for image, dockerfile in zip(cabs,dockerfiles):
        if image not in built:
            docker.build(image,
                         dockerfile,
                         build_args=build_args, args=no_cache)

            log.log_image(image, dockerfile, replace=True, cab=True)
            log.write()
            built.append(image)


def get_cabs(logfile):
    log = logger.StimelaLogger(logfile)
    cabs_ = log.read()['images']
    
    # Remove images that are not cabs
    for key in cabs_.keys():
        if not cabs_[key]['CAB']:
            del cabs_[key]

    return cabs_


def info(cabdir, header=False):
    """ prints out help information about a cab """

    # First check if cab exists
    pfile = "{}/parameters.json".format(cabdir)
    if not os.path.exists(pfile):
        raise RuntimeError("Cab could not be found at : {}".format(cabdir))
    # Get cab info
    cab_definition = cab.CabDefinition(parameter_file=pfile)
    cab_definition.display(header)


def cabs(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv = ' ' + arg

    parser = ArgumentParser(description='List executor (a.k.a cab) images')
    parser.add_argument("-i", "--cab-doc", 
        help="Will display document about the specified cab. For example, \
to get help on the 'cleanmask cab' run 'stimela cabs --cab-doc cleanmask'")

    parser.add_argument("-l", "--list", action="store_true",
            help="List cab names")

    parser.add_argument("-ls", "--list-summary", action="store_true",
            help="List cabs with a summary of the cab")

    parser.add_argument("-bl", "--build-label", default=USER,
            help="Label for build you want documentation for. See --build-label option in 'stimela help build'")

    args = parser.parse_args(argv)
    logfile = '{0:s}/{1:s}_stimela_logfile.json'.format(LOG_HOME, args.build_label)

    if args.cab_doc:
        name = '{0:s}_cab/{1:s}'.format(args.build_label, args.cab_doc)
        cabdir = "{:s}/{:s}".format(cargo.CAB_PATH, args.cab_doc)
        info(cabdir)

    elif args.list_summary:
        for val in CAB:
            cabdir = "{:s}/{:s}".format(cargo.CAB_PATH, val)
            try:
                info(cabdir, header=True)
            except IOError:
                pass
    else:
        # print them cabs
        print( ', '.join(CAB) )


def run(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Dockerized Radio Interferometric Scripting Framework.\n'
                                        'Sphesihle Makhathini <sphemakh@gmail.com>')

    add = parser.add_argument

    add("-in", "--input",
            help="Input folder")

    add("-out", "--output",
            help="Output folder")

    add("-ms", "--msdir",
            help="MS folder. MSs should be placed here. Also, empty MSs will be placed here")

    add("-j", "--ncores", type=int,
            help="Number of cores to when stimela parallesization (stimea.utils.pper function) ")

    add("script",
            help="Run script")

    add("-g", "--globals", metavar="KEY=VALUE[:TYPE]", action="append", default=[],
            help="Global variables to pass to script. The type is assumed to string unless specified")

    add("-bl", "--build-label", default=USER,
            help="Label for cab images. All cab images will be named <CAB_LABEL>_<cab name>. The default is $USER")

    args = parser.parse_args(argv)
    tag =  None

    _globals = dict(_STIMELA_INPUT=args.input, _STIMELA_OUTPUT=args.output,
                    _STIMELA_MSDIR=args.msdir,
                    CAB_TAG=tag, _STIMELA_BUILD_LABEL=args.build_label)

    nargs = len(args.globals)

    global GLOBALS

    if nargs:
        for arg in args.globals:
            if arg.find("=") > 1:
                key, value = arg.split("=")

                try:
                    value, _type = value.split(":")
                except ValueError:
                    _type = "str"

                GLOBALS[key] = eval("{:s}('{:s}')".format(_type, value))

    if args.ncores:
        utils.CPUS = args.ncores

    with open(args.script, 'r') as stdr:
        exec(stdr.read(), _globals)


def pull(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Pull docker stimela base images')

    add = parser.add_argument

    add("-im", "--image", action="append", metavar="IMAGE[:TAG]",
            help="Pull base image along with its tag (or version). Can be called multiple times")

    add("-t", "--tag",
            help="Tag")

    add("-s", "--singularity", action="store_true",
            help="Pull base images using singularity."
                 "Images will be pulled into the directory specified by the enviroment varaible, SINGULARITY_PULLFOLDER. $PWD by default")

    add("-d", "--docker", action="store_true",
            help="Pull base images using docker.")

    add("-pf", "--pull-folder", help="Images will be placed in this folder. Else, if the environmnental variable 'SINGULARITY_PULLFOLDER' is set, then images will be placed there. "
                                    "Else, images will be placed in the current directory")

    args = parser.parse_args(argv)

    if args.pull_folder:
        pull_folder = args.pull_folder
    else:
        try:
            pull_folder = os.environ["SINGULARITY_PULLFOLDER"]
        except KeyError:
            pull_folder = "."

    log = logger.StimelaLogger(LOG_FILE, jtype="docker" if args.docker else "udocker")
    images = log.read()['images']

    if args.image:
        for image in args.image:

            simage = image.replace("/", "_")
            simage = simage.replace(":", "_") + ".img"
            if args.singularity:
                singularity.pull(image, simage, directory=pull_folder)
            elif args.docker:
                docker.pull(image)
                log.log_image(image, 'pulled')
            else:
                udocker.pull(image)
                log.log_image(image, 'pulled')
    else:

        base = []
        for cab in CAB:
            image = "{:s}/{:s}".format(cargo.CAB_PATH, cab)
            base.append( utils.get_Dockerfile_base_image(image).split()[-1] )

        base = set(base)

        for image in base:
            if args.singularity:
                simage = image.replace("/", "_")
                simage = simage.replace(":", "_") + ".img"
                singularity.pull(image, simage, directory=pull_folder)
            elif args.docker:
                docker.pull(image)
                log.log_image(image, 'pulled')
            else:
               udocker.pull(image)
               log.log_image(image, 'pulled')

    log.write()


def images(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all stimela related images.')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
            help="Clear the logfile that keeps track of stimela images. This does not do anythig to the images.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('images')

    if args.clear:
        log.clear('images')
        log.write()


def containers(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all active stimela containers.')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
            help="Clear the log file that keeps track of stimela containers. This doesn't do anything to the containers.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('containers')
    if args.clear:
        log.clear('containers')
        log.write()


def ps(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='List all running stimela processes')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
            help="Clear logfile that keeps track of stimela processes. This doesn't do anything ot the processes themselves.")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log.display('processes')
    if args.clear:
        log.clear('processes')
        log.write()


def kill(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Gracefully kill stimela process(s).')

    add = parser.add_argument

    add("pid", nargs="*",
            help="Process ID")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)

    for pid in args.pid:

        found = pid in log.info['processes'].keys()

        if not found:
            print("Could not find process {0}".format(pid))
            continue

        try:
            os.kill(int(pid), signal.SIGINT)
        except OSError:
            raise OSError('Process with PID {} could not be killed'.format(pid))

        log.remove('processes', pid)
    log.write()


def clean(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg

    parser = ArgumentParser(description='Convience tools for cleaning up after stimela')
    add = parser.add_argument

    add("-ai", "--all-images", action="store_true",
        help="Remove all images pulled/built by stimela. This include CAB images")
    
    add("-ab", "--all-base", action="store_true",
        help="Remove all base images")

    add("-ac", "--all-cabs", action="store_true", 
        help="Remove all CAB images")

    add("-aC", "--all-containers", action="store_true",
        help="Stop and/or Remove all stimela containers")
    
    add("-bl", "--build-label", default=USER,
            help="Label for cab images. All cab images will be named <CAB_LABEL>_<cab name>. The default is $USER")

    args = parser.parse_args(argv)

    log = logger.StimelaLogger(LOG_FILE)
    log_cabs = logger.StimelaLogger('{0:s}/{1:s}_stimela_logfile.json'.format(LOG_HOME, 
            args.build_label))

    if args.all_images:
        images = log.info['images'].keys()
        images = log_cabs.info['images'].keys()
        for image in images:
            utils.xrun('docker', ['rmi', image])
            log.remove('images', image)
            log.write()
        
        images = log_cabs.info['images'].keys()
        for image in images:
            if log_cabs.info['images'][image]['CAB']:
                utils.xrun('docker', ['rmi', image])
                log_cabs.remove('images', image)
                log_cabs.write()

    if args.all_base:
        images = log.info['images'].keys()
        for image in images:
            if log.info['images'][image]['CAB'] is False:
                utils.xrun('docker', ['rmi', image])
                log.remove('images', image)
                log.write()

    if args.all_cabs:
        images = log_cabs.info['images'].keys()
        for image in images:
            if log_cabs.info['images'][image]['CAB']:
                utils.xrun('docker', ['rmi', image])
                log_cabs.remove('images', image)
                log_cabs.write()
        
    if args.all_containers:
        containers = log.info['containers'].keys()
        for container in containers:
            cont = docker.Container(log.info['containers'][container]['IMAGE'], container)
            try:
                status = cont.info()['State']['Status'].lower()
            except:
                print('Could not inspect container {}. It probably doesn\'t exist, will remove it from log'.format(container))
                status = "no there"

            if status == 'running':
            # Kill the container instead of stopping it, so that effect can be felt py parent process
                utils.xrun('docker', ['kill', container])
                cont.remove()
            elif status in ['exited', 'dead']:
                cont.remove()

            log.remove('containers', container)
            log.write()

def main(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit(): argv[i] = ' ' + arg


    parser = ArgumentParser(description='Stimela: Dockerized Radio Interferometric Scripting Framework. '
                            '|n version {:s} |n install path {:s} |n '
                            'Sphesihle Makhathini <sphemakh@gmail.com>'.format(__version__,
                                                                               os.path.dirname(__file__)),
                            formatter_class=MultilineFormatter,
                            add_help=False)

    add = parser.add_argument

    add("-h", "--help",  action="store_true",
            help="Print help message and exit")

    add("-v","--version", action='version',
            version='{:s} version {:s}'.format(parser.prog, __version__))

    add("command", nargs="*", metavar="command [options]",
            help="Stimela command to execute. For example, 'stimela help run'")

    options = []
    commands = dict(pull=pull, build=build, run=run,
                    images=images, cabs=cabs, ps=ps,
                    containers=containers, kill=kill,
                    clean=clean)

    command = "failure"

    for cmd in commands:
        if cmd in argv:
            command = cmd

            index = argv.index(cmd)
            options = argv[index+1:]
            argv = argv[:index+1]

    args = parser.parse_args(argv)

    # Command is help and no other commands following
    try:
        main_help = (args.command[0] == "help" and len(args.command) == 1)
    except IndexError:
        main_help = True

    if args.help or main_help:
        parser.print_help()

        print ("""
Run a command. These can be:

help    : Prints out a help message about other commands
build   : Build a set of stimela images
pull    : pull a stimela base images
run     : Run a stimela script
images  : List stimela images
cabs    : Manage cab images
ps      : List running stimela scripts
kill    : Gracefully kill runing stimela process
clean   : Clean up tools for stimela

""")

        sys.exit(0)

    # Separate commands into command and arguments
    cmd, argv = args.command[0], args.command[1:]

    # If we've got past the if statement above, and help
    # is the command then assume that help on a command
    # is requested
    if cmd == "help":
        # Request help on the sub-command
        cmd, argv = argv[0], ["-h"]
    else:
        argv = options

    # Get the function to execute for the command
    try:
        _cmd = commands[cmd]
    except KeyError:
        raise KeyError("Command '{:s}' not recognized "
                         "Run : 'stimela help' for help".format(cmd))

    # Invoke the command
    _cmd(argv)



import os
import sys
import logging
import Crasa.Crasa as crasa

sys.path.append("/scratch/stimela")
import utils

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)

args = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    args[name] = value

task = crasa.CasaTask(cab["binary"], **args)
task.run()



import os
import sys

sys.path.append('/scratch/stimela')
import utils


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

cab = utils.readJson(CONFIG)
mslist = []
field = []

args = []
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value in [False, None]:
        continue
    if name == "primary-beam":
        value = "'{}'".format(value)
    if name == 'pa-range' and hasattr(value, '__iter__'):
        value = ','.join(value)
    if value is True:
        value = ""
    if name == 'pa-from-ms' and hasattr(value, '__iter__'):
        mslist = value
        continue
    if name == 'field-id'and hasattr(value, '__iter__'):
        field = value
        continue

    # Positional arguments
    if name == 'input-skymodel':
        inlsm = value
        continue
    elif name == 'output-skymodel':
        outlsm = value
        continue

    args.append( '{0}{1} {2}'.format(cab['prefix'], name, value) )

if mslist:
    if len(field)==0:
        field = [0]*len(mslist)
    pa_from_ms = ','.join(['{0}:{1}'.format(ms, i) for ms,i in zip(mslist, field)])
    args.append('--pa-from-ms {}'.format(pa_from_ms))

utils.xrun(cab['binary'], args + [inlsm, outlsm])



import os
import sys
import re
import pyfits

sys.path.append('/scratch/stimela')
import utils

CONFIG = os.environ['CONFIG']
INPUT = os.environ['INPUT']
OUTPUT = os.environ['OUTPUT']
MSDIR = os.environ['MSDIR']

cab = utils.readJson(CONFIG)
params = cab['parameters']
args = []

# new interface as of WSCLEAN v2.5
# make the transition transparent

npix = filter(lambda x: x['name'] == 'size', params)[0]['value']

if isinstance(npix, int):
    npix = [npix, npix]
elif isinstance(npix, list) and len(npix) == 1:
    npix = npix * 2
elif not (isinstance(npix, list) and
          all([isinstance(px, int) for px in npix]) and
          len(npix) == 2) :
    raise ValueError("Npix only accepts single int or list[2] of int")

trdefault = int(max(npix[0], npix[1]) * 0.75)
trim = filter(lambda x: x['name'] == 'trim', params)[0]['value']

if trim is None:
    trim = [trdefault, trdefault]
elif isinstance(trim, int):
    trim = [trim, trim]
elif isinstance(trim, list) and len(trim) == 1:
    trim = trim * 2
elif not (isinstance(trim, list) and
          all([isinstance(t, int) for t in trim]) and
          len(npix) == 2) :
    raise ValueError("trim only accepts single int or list[2] of int")
pad = max(npix[0], npix[1]) / float(min(trim[0], trim[1]))
trindex = filter(lambda x: x['name'] == 'trim', params)[0]
params.remove(trindex)
padding = filter(lambda x: x['name'] == 'padding', params)
if not padding:
    filter(lambda x: x['name'] == 'size', params)[0]["value"] = trim  # npix is now the unpadded size
    params.append({'name': 'padding', 'value': pad})  # replace with 'padding' argument

for param in params:
    name = param['name']
    value = param['value']

    if name == 'msname':
        if isinstance(value, str):
            mslist = value
        else:
            mslist = ' '.join(value)
        continue

    if value in [None, False]:
        continue
    # only versions > 2.0 have this 
    if name in ["spws"]:
        continue

    if name == 'scale':
        if isinstance(value, (int, float)):
            value = '{0}asec'.format(value)

    if name == 'datacolumn':
        name = 'data-column' # new interface as of WSCLEAN 2.5 - here for compat

    if name in 'size trim nwlayers-for-size beam-shape channel-range interval'.split():
        if isinstance(value, int):
            value = '{0} {0}'.format(value)
        elif hasattr(value, '__iter__'):
            if len(value) == 1:
                value.append( value[0])
            value = ' '.join(map(str, value))

    if name in 'spws multiscale-scales pol'.split():
        if hasattr(value, '__iter__'):
            value = ','.join(map(str, value))

    if value is True:
        arg = '{0}{1}'.format(cab['prefix'], name)
    else:
        arg = '{0}{1} {2}'.format(cab['prefix'], name, value)

    args.append(arg)

if '{0}auto-threshold'.format(cab['prefix']) not in args:
    removed = False
    for item1 in args:
        if 'noise-image' in item1:
            noise_image = item1.split('{0}noise-image '.format(cab['prefix']))[-1]
            args.remove('{0}noise-image {1}'.format(cab['prefix'], noise_image))
            noise_hdu = pyfits.open(noise_image)
            noise_data = noise_hdu[0].data
            noise_std = noise_data.std()
            noise_hdu.close()
            for item2 in args:
                if 'noise-sigma' in item2:
                    noise_sigma = item2.split('{0}noise-sigma '.format(cab['prefix']))[-1]
                    args.remove('{0}noise-sigma {1}'.format(cab['prefix'], noise_sigma))
                    removed = True
                    threshold = float(noise_sigma)*noise_std
                    for item3 in args:
                        if '{0}threshold'.format(cab['prefix']) in item3:
                            args.remove(item3)
                    args.append('{0}threshold {1}'.format(cab['prefix'], threshold))
    if not removed:
        args.remove('{0}noise-sigma 3.0'.format(cab['prefix']))
else:
    args.remove('{0}noise-sigma 3.0'.format(cab['prefix']))

utils.xrun(cab['binary'], args + [mslist])



import os
import sys
import json
import yaml
import time
import subprocess
from io import StringIO
import codecs
from datetime import datetime

class StimelaLogger(object):
    def __init__(self, lfile, jtype="udocker"):

        self.lfile = lfile
        # Create file if it does not exist
        if not os.path.exists(self.lfile):
            with open(lfile, 'w') as wstd:
                wstd.write('{}')
            
        self.info = self.read(lfile)
        # First make sure that all fields are
        # initialised. Initialise if not so
        changed = False
        for item in ['images', 'containers', 'processes']:
            if self.info.get(item, None) is None:
                self.info[item] = {}
                changed = True
        if changed:
            self.write()


        self.jtype = jtype

    def _inspect(self, name):
        
        output = subprocess.check_output("{0:s} inspect {1:s}".format(self.jtype, name), shell=True).decode()
        if self.jtype == "docker":
            output_file = StringIO(output[3:-3])
        else:
            output_file = StringIO(output)
        jdict = yaml.safe_load(output_file)
        output_file.close()

        return jdict

    def log_image(self, name, image_dir, replace=False, cab=False):
        info = self._inspect(name)
        if self.jtype == "docker":
            if name not in self.info['images'].keys() or replace:
                self.info['images'][name] = {
                    'TIME'      :   info['Created'].split('.')[0].replace('Z', '0'),
                    'ID'        :   info['Id'].split(':')[1],
                    'CAB'       :   cab,
                    'DIR'       :   image_dir, 
                }
            else:
                print('Image {0} has already been logged.'.format(name))
        else:
            if name not in self.info['images'].keys() or replace:
                self.info['images'][name] = {
                    'TIME'      :   info['created'].split('.')[0].replace('Z', '0'),
                    'ID'        :   info['id'],
                    'CAB'       :   cab,
                    'DIR'       :   image_dir, 
                }
            else:
                print('Image {0} has already been logged.'.format(name))


    def log_container(self, name):
        info = self._inspect(name)

        if self.jtype == "docker":
            if name not in self.info['containers'].keys():
                self.info['containers'][name] = {
                    'TIME'      :   info['Created'].split('.')[0].replace('Z', '0'),
                    'IMAGE'     :   info['Config']['Image'],
                    'ID'        :   info['Id'],
                }
            else:
                print('contaier {0} has already been logged.'.format(name))
        else:
            if name not in self.info['containers'].keys():
                self.info['containers'][name] = {
                    'TIME'      :   info['created'].split('.')[0].replace('Z', '0'),
                    'IMAGE'     :   info['config']['Image'],
                    'ID'        :   info['id'],
                }
            else:
                print('contaier {0} has already been logged.'.format(name))


    def log_process(self, pid, name):
        pid = str(pid)
        timestamp = time.strftime('%Y-%m-%dT%H:%M:%S', time.gmtime())
        if pid not in self.info['processes'].keys():
            self.info['processes'][pid] = {
                'NAME'  :   name,
                'TIME'  :   timestamp,
            }
        else:
            print('PID {0} has already been logged.'.format(pid))


    def remove(self, ltype, name):
        try:
            self.info[ltype].pop(str(name))
        except:
            print('WARNING:: Could not remove object \'{0}:{1}\' from logger'.format(ltype, name))
    
    
    def read(self, lfile=None):
        try:
            with open(lfile or self.lfile) as _std:
                jdict = yaml.safe_load(_std)
        except IOError:
            return {}

        return jdict


    def write(self, lfile=None):
        with codecs.open(lfile or self.lfile, 'w', 'utf8') as std:
            std.write(json.dumps(self.info, ensure_ascii=False, indent=4))


    def clear(self, ltype):
        self.info[ltype] = {}


    def display(self, ltype):
        things = sorted(self.info[ltype].items(), key=lambda a: a[1]['TIME'])
        if ltype == 'images':
            print('{0:<36}      {1:<24}     {2:<24}'.format('IMAGE', 'ID', 'CREATED/PULLED'))
            for name, thing in things:
                print('{0:<36}      {1:<24}     {2:<24}'.format(name, thing['ID'][:8], thing['TIME']))

        if ltype == 'containers':
            print('{0:<36}      {1:<24}     {2:<24}     {3:<24}      {4:<24}'.format('CONTAINER','CAB IMAGE', 'ID', 'STARTED', 'UPTIME'))
            for name, thing in things:
                started = datetime.strptime(thing['TIME'], '%Y-%m-%dT%H:%M:%S')
                finished = datetime.utcnow()
                mins = (finished - started).total_seconds()/60
                hours, mins = divmod(mins, 60)
                mins, secs = divmod(mins*60, 60)

                uptime = '{0:d}:{1:d}:{2:.2f}'.format(int(hours), int(mins), secs)
                print('{0:<36}      {1:<24}     {2:<24}     {3:<24}      {4:<24}'.format(name, thing['IMAGE'], thing['ID'][:8], thing['TIME'], uptime))

        if ltype == 'processes':
            print('{0:<36}      {1:<24}     {2:<24}     {3:24}'.format('PID','NAME', 'STARTED', 'UPTIME'))
            for name, thing in things:
                started = datetime.strptime(thing['TIME'], '%Y-%m-%dT%H:%M:%S')
                finished = datetime.utcnow()
                mins = (finished - started).total_seconds()/60
                hours, mins = divmod(mins, 60)
                mins, secs = divmod(mins*60, 60)

                uptime = '{0:d}:{1:d}:{2:.2f}'.format(int(hours), int(mins), secs)
                print('{0:<36}      {1:<24}     {2:<24}     {3:24}'.format(name,thing['NAME'], thing['TIME'], uptime))



#!/usr/bin/env python

import os
import sys
from setuptools import setup


requirements = ["pyyaml", 
                "nose>=1.3.7",
                "future",
               ],

if sys.version_info <= (3, 0):
    requirements += [
                "udocker-stimela",
               ],

PACKAGE_NAME = "stimela"
__version__ = "1.1.2"

setup(name = PACKAGE_NAME,
    version = __version__,
    description = "Dockerized radio interferometry scripting framework",
    author = "Sphesihle Makhathini",
    author_email = "sphemakh@gmail.com",
    url = "https://github.com/sphemakh/Stimela",
    packages = ["stimela","stimela_misc", "stimela/cargo","stimela/utils", "stimela/cargo/cab"],
    package_data = { "stimela/cargo" : [
                                   "cab/*/Dockerfile",
                                   "base/*/Dockerfile",
                                   "cab/*/src/*.py",
                                   "cab/*/src/*.sh",
                                   "cab/*/src/*.json",
                                   "cab/*/xvfb.init.d",
                                   "cab/*/parameters.json",
                                   "cab/*/src/tdlconf.profiles",
                                   "cab/singularity_run",
                                   ]},
    install_requires = requirements,
    scripts = ["bin/" + i for i in os.listdir("bin")],
    classifiers = [],
     )



import os
import sys

sys.path.append("/scratch/stimela")
import utils


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTDIR = os.environ["OUTPUT"]

cab = utils.readJson(CONFIG)
args = []
for param in cab['parameters']:
    value = param['value']
    name = param['name']

    if value in [None, False]:
        continue
    elif value is True:
        value = ""
    elif name == 'hdf5files':
        files = value
        continue

    args += ['{0}{1} {2}'.format(cab['prefix'], name, value)]

utils.xrun("h5toms.py", args+files )
